import time

import arjun.core.config as mem

from arjun.core.colors import bad


def connection_refused():
	"""
	checks if a request should be retried if the server refused connection
	returns str
	"""
	if mem.var['stable']:
		print('%s Hit rate limit, stabilizing the connection' % bad)
		mem.var['kill'] = False
		time.sleep(30)
		return 'retry'
	print('%s Target has rate limiting in place, please use --stable switch' % bad)
	return 'kill'


def error_handler(response, factors):
	"""
	decides what to do after performing a HTTP request
		'ok': continue normally
		'retry': retry this request
		'kill': stop processing this target
	returns str
	"""
	if type(response) != str and response.status_code in (400, 413, 418, 429, 503):
		if not mem.var['healthy_url']:
			return 'ok'
		if response.status_code == 503:
			mem.var['kill'] = True
			print('%s Target is unable to process requests, try --stable switch' % bad)
			return 'kill'
		elif response.status_code in (429, 418):
			print('%s Target has a rate limit in place, try --stable switch' % bad)
			return 'kill'
		else:
			if factors['same_code'] != response.status_code:
				mem.var['bad_req_count'] = mem.var.get('bad_req_count', 0) + 1
				if mem.var['bad_req_count'] > 20:
					mem.var['kill'] = True
					print('%s Server received a bad request. Try decreasing the chunk size with -c option' % bad)
					return 'kill'
			else:
				return 'ok'
	else:
		if 'Timeout' in response:
			if mem.var['timeout'] > 20:
				mem.var['kill'] = True
				print('%s Connection timed out, unable to increase timeout further' % bad)
				print('%s Target might have a rate limit in place, try --stable switch' % bad)
				return 'kill'
			else:
				print('%s Connection timed out, increased timeout by 5 seconds' % bad)
				mem.var['timeout'] += 5
				return 'retry'
		elif 'ConnectionRefused' in response:
			return connection_refused()
		elif type(response) == str:
			if '\'' in response:
				print('%s Encountered an error: %s' % (bad, response.split('\'')[1]))
			return 'kill'
	return 'ok'



import json
import time
import random
import requests
import warnings

import arjun.core.config as mem

from ratelimit import limits, sleep_and_retry
from arjun.core.utils import dict_to_xml

warnings.filterwarnings('ignore') # Disable SSL related warnings


@sleep_and_retry
@limits(calls=mem.var['rate_limit'], period=1)
def requester(request, payload={}):
    """
    central function for making http requests
    returns str on error otherwise response object of requests library
    """
    if request.get('include') and len(request.get('include', '')) != 0:
        payload.update(request['include'])
    if mem.var['stable']:
        mem.var['delay'] = random.choice(range(3, 10))
    time.sleep(mem.var['delay'])
    url = request['url']
    if mem.var['kill']:
        return 'killed'
    try:
        if request['method'] == 'GET':
            response = requests.get(url,
                params=payload,
                headers=request['headers'],
                verify=False,
                allow_redirects=False,
                timeout=mem.var['timeout'],
            )
        elif request['method'] == 'JSON':
            request['headers']['Content-Type'] = 'application/json'
            if mem.var['include'] and '$arjun$' in mem.var['include']:
                payload = mem.var['include'].replace('$arjun$',
                    json.dumps(payload).rstrip('}').lstrip('{'))
                response = requests.post(url,
                    data=payload,
                    headers=request['headers'],
                    verify=False,
                    allow_redirects=False,
                    timeout=mem.var['timeout'],
                )
            else:
                response = requests.post(url,
                    json=payload,
                    headers=request['headers'],
                    verify=False,
                    allow_redirects=False,
                    timeout=mem.var['timeout'],
                )
        elif request['method'] == 'XML':
            request['headers']['Content-Type'] = 'application/xml'
            payload = mem.var['include'].replace('$arjun$',
                dict_to_xml(payload))
            response = requests.post(url,
                data=payload,
                headers=request['headers'],
                verify=False,
                allow_redirects=False,
                timeout=mem.var['timeout'],
            )
        else:
            response = requests.post(url,
                data=payload,
                headers=request['headers'],
                verify=False,
                allow_redirects=False,
                timeout=mem.var['timeout'],
            )
        return response
    except Exception as e:
        return str(e)



import requests

from urllib.parse import urlparse


def wayback(host, page):
	payload = {
		'url': host,
		'matchType': 'host',
		'collapse': 'urlkey',
		'fl': 'original',
		'page': page,
		'limit': 10000
	}
	headers = {
		'User-Agent': 'Mozilla'
	}
	try:
		these_params = set()
		response = requests.get(
			'http://web.archive.org/cdx/search?filter=mimetype:text/html&filter=statuscode:200',
			params=payload,
			headers=headers,
			verify=False
		).text
		if not response:
			return (these_params, False, 'wayback')
		urls = filter(None, response.split('\n'))
		for url in urls:
			for param in urlparse(url).query.split('&'):
				these_params.add(param.split('=')[0])
		return (these_params, True, 'wayback')
	except requests.exceptions.ConnectionError:
		return (these_params, False, 'wayback')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from arjun.core.colors import green, end, info, bad, good, run, res

import argparse
import json

from urllib.parse import urlparse
import arjun.core.config as mem
from arjun.core.exporter import exporter
from arjun.core.anomaly import define, compare
from arjun.core.utils import fetch_params, stable_request, random_str, slicer, confirm, populate, reader, nullify, prepare_requests, compatible_path

from arjun.plugins.heuristic import heuristic
from arjun.plugins.wl import detect_casing, covert_to_case

arjun_dir = compatible_path(mem.__file__.replace(compatible_path('/core/config.py'), ''))

parser = argparse.ArgumentParser() # defines the parser
# Arguments that can be supplied
parser.add_argument('-u', help='Target URL', dest='url')
parser.add_argument('-o', '-oJ', help='Path for json output file.', dest='json_file')
parser.add_argument('-oT', help='Path for text output file.', dest='text_file')
parser.add_argument('-oB', help='Output to Burp Suite Proxy. Default is 127.0.0.1:8080.', dest='burp_proxy', nargs='?', const='127.0.0.1:8080')
parser.add_argument('-d', help='Delay between requests in seconds. (default: 0)', dest='delay', type=float, default=0)
parser.add_argument('-t', help='Number of concurrent threads. (default: 5)', dest='threads', type=int, default=5)
parser.add_argument('-w', help='Wordlist file path. (default: {arjundir}/db/large.txt)', dest='wordlist', default=arjun_dir+'/db/large.txt')
parser.add_argument('-m', help='Request method to use: GET/POST/XML/JSON. (default: GET)', dest='method', default='GET')
parser.add_argument('-i', help='Import target URLs from file.', dest='import_file', nargs='?', const=True)
parser.add_argument('-T', help='HTTP request timeout in seconds. (default: 15)', dest='timeout', type=float, default=15)
parser.add_argument('-c', help='Chunk size. The number of parameters to be sent at once', type=int, dest='chunks', default=250)
parser.add_argument('-q', help='Quiet mode. No output.', dest='quiet', action='store_true')
parser.add_argument('--rate-limit', help='Max number of requests to be sent out per second (default: 9999)', dest='rate_limit', type=int, default=9999)
parser.add_argument('--headers', help='Add headers. Separate multiple headers with a new line.', dest='headers', nargs='?', const=True)
parser.add_argument('--passive', help='Collect parameter names from passive sources like wayback, commoncrawl and otx.', dest='passive', nargs='?', const='-')
parser.add_argument('--stable', help='Prefer stability over speed.', dest='stable', action='store_true')
parser.add_argument('--include', help='Include this data in every request.', dest='include', default={})
parser.add_argument('--disable-redirects', help='disable redirects', dest='disable_redirects', action='store_true')
parser.add_argument('--casing', help='casing style for params e.g. like_this, likeThis, likethis', dest='casing')
args = parser.parse_args() # arguments to be parsed

if args.quiet:
    print = nullify

print('''%s    _
   /_| _ '
  (  |/ /(//) v%s
      _/      %s
''' % (green, __import__('arjun').__version__, end))

try:
    from concurrent.futures import ThreadPoolExecutor, as_completed
except ImportError:
    print('%s Please use Python > 3.2 to run Arjun.' % bad)
    quit()

mem.var = vars(args)

mem.var['method'] = mem.var['method'].upper()

if mem.var['method'] != 'GET':
    mem.var['chunks'] = 500

if mem.var['stable'] or mem.var['delay']:
    mem.var['threads'] = 1
if mem.var['wordlist'] in ('large', 'medium', 'small'):
    mem.var['wordlist'] = f'{arjun_dir}/db/{mem.var["wordlist"]}.txt'

try:
    wordlist_file = arjun_dir + '/db/small.txt' if args.wordlist == 'small' else args.wordlist
    wordlist_file = compatible_path(wordlist_file)
    wordlist = set(reader(wordlist_file, mode='lines'))
    if mem.var['passive']:
        host = mem.var['passive']
        if host == '-':
            host = urlparse(args.url).netloc
        print('%s Collecting parameter names from passive sources for %s, it may take a while' % (run, host))
        passive_params = fetch_params(host)
        wordlist.update(passive_params)
        print('%s Collected %s parameters, added to the wordlist' % (info, len(passive_params)))
    if args.casing:
        delimiter, casing = detect_casing(args.casing)
        wordlist = [covert_to_case(word, delimiter, casing) for word in wordlist]
    else:
        wordlist = list(wordlist)
except FileNotFoundError:
    exit('%s The specified file for parameters doesn\'t exist' % bad)

if len(wordlist) < mem.var['chunks']:
    mem.var['chunks'] = int(len(wordlist)/2)

if not args.url and not args.import_file:
    exit('%s No target(s) specified' % bad)

from arjun.core.requester import requester
from arjun.core.bruter import bruter

def narrower(request, factors, param_groups):
    """
    takes a list of parameters and narrows it down to parameters that cause anomalies
    returns list
    """
    anomalous_params = []
    threadpool = ThreadPoolExecutor(max_workers=mem.var['threads'])
    futures = (threadpool.submit(bruter, request, factors, params) for params in param_groups)
    for i, result in enumerate(as_completed(futures)):
        if result.result():
            anomalous_params.extend(slicer(result.result()))
        if mem.var['kill']:
            return anomalous_params
        print('%s Processing chunks: %i/%-6i' % (info, i + 1, len(param_groups)), end='\r')
    return anomalous_params


def initialize(request, wordlist, single_url=False):
    """
    handles parameter finding process for a single request object
    returns 'skipped' (on error), list on success
    """
    url = request['url']
    if not url.startswith('http'):
        print('%s %s is not a valid URL' % (bad, url))
        return 'skipped'
    print('%s Probing the target for stability' % run)
    request['url'] = stable_request(url, request['headers'])
    mem.var['healthy_url'] = True
    if not request['url']:
        return 'skipped'
    else:
        fuzz = "z" + random_str(6)
        response_1 = requester(request, {fuzz[:-1]: fuzz[::-1][:-1]})
        mem.var['healthy_url'] = response_1.status_code not in (400, 413, 418, 429, 503)
        if not mem.var['healthy_url']:
            print('%s Target returned HTTP %i, this may cause problems.' % (bad, request.status_code))
        if single_url:
            print('%s Analysing HTTP response for anomalies' % run)
        response_2 = requester(request, {fuzz[:-1]: fuzz[::-1][:-1]})
        if type(response_1) == str or type(response_2) == str:
            return 'skipped'

        # params from response must be extracted before factors but displayed later
        found, words_exist = heuristic(response_1, wordlist)

        factors = define(response_1, response_2, fuzz, fuzz[::-1], wordlist)
        zzuf = "z" + random_str(6)
        response_3 = requester(request, {zzuf[:-1]: zzuf[::-1][:-1]})
        while True:
            reason = compare(response_3, factors, {zzuf[:-1]: zzuf[::-1][:-1]})[2]
            if not reason:
                break
            factors[reason] = None
        if found:
            num = len(found)
            if words_exist:
                print('%s Extracted %i parameters from response for testing' % (good, num))
            else:
                s = 's' if num > 1 else ''
                print('%s Extracted %i parameter%s from response for testing: %s' % (good, num, s, ', '.join(found)))
        if single_url:
            print('%s Logicforcing the URL endpoint' % run)
        populated = populate(wordlist)
        with open(f'{arjun_dir}/db/special.json', 'r') as f:
            populated.update(json.load(f))
        param_groups = slicer(populated, int(len(wordlist)/mem.var['chunks']))
        prev_chunk_count = len(param_groups)
        last_params = []
        while True:
            param_groups = narrower(request, factors, param_groups)
            if len(param_groups) > prev_chunk_count:
                response_3 = requester(request, {zzuf[:-1]: zzuf[::-1][:-1]})
                if compare(response_3, factors, {zzuf[:-1]: zzuf[::-1][:-1]})[0] != '':
                    print('%s Webpage is returning different content on each request. Skipping.' % bad)
                    return []
            if mem.var['kill']:
                return 'skipped'
            param_groups = confirm(param_groups, last_params)
            prev_chunk_count = len(param_groups)
            if not param_groups:
                break
        confirmed_params = []
        for param in last_params:
            reason = bruter(request, factors, param, mode='verify')
            if reason:
                name = list(param.keys())[0]
                confirmed_params.append(name)
                if single_url:
                    print('%s parameter detected: %s, based on: %s' % (res, name, reason))
        return confirmed_params


def main():
    requests = prepare_requests(args)

    final_result = {}
    is_single = False if args.import_file else True

    try:
        mem.var['kill'] = False
        count = 0
        for request in requests:
            url = request['url']
            print('%s Scanning %d/%d: %s' % (run, count, len(requests), url))
            these_params = initialize(request, wordlist, single_url=is_single)
            count += 1
            mem.var['kill'] = False
            mem.var['bad_req_count'] = 0
            if these_params == 'skipped':
                print('%s Skipped %s due to errors' % (bad, url))
            elif these_params:
                final_result[url] = {}
                final_result[url]['params'] = these_params
                final_result[url]['method'] = request['method']
                final_result[url]['headers'] = request['headers']
                exporter(final_result)
                print('%s Parameters found: %-4s\n' % (good, ', '.join(final_result[url]['params'])))
                if not mem.var['json_file']:
                    final_result = {}
                    continue
            else:
                print('%s No parameters were discovered.\n' % info)
    except KeyboardInterrupt:
        exit()


if __name__ == '__main__':
    main()



import re

from arjun.core.colors import info
import arjun.core.config as mem
from arjun.core.utils import extract_js

# TODO: for map keys, javascript tolerates { param: "value" }
re_words = re.compile(r'[A-Za-z][A-Za-z0-9_]*')
re_not_junk = re.compile(r'^[A-Za-z0-9_]+$')
re_inputs = re.compile(r'''(?i)<(?:input|textarea)[^>]+?(?:id|name)=["']?([^"'\s>]+)''')
re_empty_vars = re.compile(r'''(?:[;\n]|\bvar|\blet)(\w+)\s*=\s*(?:['"`]{1,2}|true|false|null)''')
re_map_keys = re.compile(r'''['"](\w+?)['"]\s*:\s*['"`]''')


def is_not_junk(param):
    return (re_not_junk.match(param) is not None)


def heuristic(raw_response, wordlist):
    words_exist = False
    potential_params = []

    headers, response = raw_response.headers, raw_response.text
    if headers.get('content-type', '').startswith(('application/json', 'text/plain')):
        if len(response) < 200:
            if ('required' or 'missing' or 'not found' or 'requires') in response.lower() and ('param' or 'parameter' or 'field') in response.lower():
                if not mem.var['quiet']:
                    print('%s The endpoint seems to require certain parameters to function. Check the response and use the --include option appropriately for better results.' % info)
            words_exist = True
            potential_params = re_words.findall(response)
    # Parse Inputs
    input_names = re_inputs.findall(response)
    potential_params += input_names

    # Parse Scripts
    for script in extract_js(response):
        empty_vars = re_empty_vars.findall(script)
        potential_params += empty_vars

        map_keys = re_map_keys.findall(script)
        potential_params += map_keys

    if len(potential_params) == 0:
        return [], words_exist

    found = set()
    for word in potential_params:
        if is_not_junk(word) and (word not in found):
            found.add(word)

            if word in wordlist:
                wordlist.remove(word)
            wordlist.insert(0, word)

    return list(found), words_exist



def detect_casing(string):
    """Detect the casing style and delimiter of given string."""
    delimiter = ""
    casing = ""

    if string.islower():
        casing = "l"
    elif string.isupper():
        casing = "u"
    else:
        casing = casing = "c" if string[0].islower() else "p"

    if "-" in string:
        delimiter = "-"
    elif "_" in string:
        delimiter = "_"
    elif "." in string:
        delimiter = "."

    return delimiter, casing


def transform(parts, delimiter, casing):
    """Combine list of strings to form a string with given casing style."""
    if len(parts) == 1:
        if casing == "l":
            return parts[0].lower()
        elif casing == "u":
            return parts[0].upper()
        return parts[0]

    result = []
    for i, part in enumerate(parts):
        if casing == "l":
            transformed = part.lower()
        elif casing == "u":
            transformed = part.upper()
        elif casing == "c":
            if i == 0:
                transformed = part.lower()
            else:
                transformed = part.lower().title()
        else:  # casing == "p"
            transformed = part.lower().title()

        result.append(transformed)

    return delimiter.join(result)


def handle(text):
    """Break down a string into array of 'words'."""
    if "-" in text:
        return text.split("-")
    elif "_" in text:
        return text.split("_")
    elif "." in text:
        return text.split(".")

    if not text.islower() and not text.isupper():
        parts = []
        temp = ""
        for char in text:
            if not char.isupper():
                temp += char
            else:
                if temp:
                    parts.append(temp)
                temp = char
        if temp:
            parts.append(temp)
        return parts

    return [text]


def covert_to_case(string, delimiter, casing):
    """Process input stream and write transformed text to output stream."""
    parts = handle(string)
    return transform(parts, delimiter, casing)



import sys

colors = True # Output should be colored
machine = sys.platform # Detecting the os of current system
if machine.lower().startswith(('os', 'win', 'darwin', 'ios')):
    colors = False # Colors shouldn't be displayed in mac & windows
if not colors:
    white = green = red = yellow = end = back = info = que = bad = good = run = res = ''
else:
    white = '\033[97m'
    green = '\033[92m'
    red = '\033[91m'
    yellow = '\033[93m'
    end = '\033[0m'
    back = '\033[7;91m'
    info = '\033[1;93m[!]\033[0m'
    que = '\033[1;94m[?]\033[0m'
    bad = '\033[1;91m[-]\033[0m'
    good = '\033[1;32m[+]\033[0m'
    run = '\033[1;97m[*]\033[0m'
    res = '\033[1;92m[✓]\033[0m'



import requests

from urllib.parse import urlparse


def otx(host, page):
	these_params = set()
	data = requests.get('https://otx.alienvault.com/api/v1/indicators/hostname/%s/url_list?limit=50&page=%d' % (host, page+1), verify=False).json()
	if 'url_list' not in data:
		return (these_params, False, 'otx')
	for obj in data['url_list']:
		for param in urlparse(obj['url']).query.split('&'):
			these_params.add(param.split('=')[0])
	return (these_params, data['has_next'], 'otx')



import arjun.core.config as mem

from arjun.core.anomaly import compare
from arjun.core.requester import requester
from arjun.core.error_handler import error_handler


def bruter(request, factors, params, mode='bruteforce'):
    """
    returns anomaly detection result for a chunk of parameters
    returns list
    """
    if mem.var['kill']:
        return []
    response = requester(request, params)
    conclusion = error_handler(response, factors)
    if conclusion == 'retry':
        return bruter(request, factors, params, mode=mode)
    elif conclusion == 'kill':
        mem.var['kill'] = True
        return []
    comparison_result = compare(response, factors, params)
    if mode == 'verify':
        return comparison_result[0]
    return comparison_result[1]



import re
import requests

import arjun.core.config as mem

from urllib.parse import urlparse
from arjun.core.utils import diff_map, remove_tags


def define(response_1, response_2, param, value, wordlist):
    """
    defines a rule list for detecting anomalies by comparing two HTTP response
    returns dict
    """
    factors = {
        'same_code': None, # if http status code is same, contains that code
        'same_body': None, # if http body is same, contains that body
        'same_plaintext': None, # if http body isn't same but is same after removing html, contains that non-html text
        'lines_num': None, # if number of lines in http body is same, contains that number
        'lines_diff': None, # if http-body or plaintext aren't and there are more than two lines, contain which lines are same
        'same_headers': None, # if the headers are same, contains those headers
        'same_redirect': None, # if both requests redirect in similar manner, contains that redirection
        'param_missing': None, # if param name is missing from the body, contains words that are already there
        'value_missing': None # contains whether param value is missing from the body
    }
    if type(response_1) == type(response_2) == requests.models.Response:
        body_1, body_2 = response_1.text, response_2.text
        if response_1.status_code == response_2.status_code:
            factors['same_code'] = response_1.status_code
        if response_1.headers.keys() == response_2.headers.keys():
            factors['same_headers'] = list(response_1.headers.keys())
            factors['same_headers'].sort()
        if mem.var['disable_redirects']:
            if response_1.headers.get('Location', '') == response_2.headers.get('Location', ''):
                factors['same_redirect'] = urlparse(response_1.headers.get('Location', '')).path
        elif urlparse(response_1.url).path == urlparse(response_2.url).path:
            factors['same_redirect'] = urlparse(response_1.url).path
        else:
            factors['same_redirect'] = ''
        if response_1.text == response_2.text:
            factors['same_body'] = response_1.text
        elif response_1.text.count('\n') == response_2.text.count('\n'):
            factors['lines_num'] = response_1.text.count('\n')
        elif remove_tags(body_1) == remove_tags(body_2):
            factors['same_plaintext'] = remove_tags(body_1)
        elif body_1 and body_2 and body_1.count('\\n') == body_2.count('\\n'):
            factors['lines_diff'] = diff_map(body_1, body_2)
        if param not in response_2.text:
            factors['param_missing'] = [word for word in wordlist if word in response_2.text]
        if value not in response_2.text:
            factors['value_missing'] = True
    return factors


def compare(response, factors, params):
    """
    detects anomalies by comparing a HTTP response against a rule list
    returns string, list (anomaly, list of parameters that caused it)
    """
    if response == '' or type(response) == str:
        return ('', [], '')
    these_headers = list(response.headers.keys())
    these_headers.sort()
    if factors['same_code'] is not None and response.status_code != factors['same_code']:
        return ('http code', params, 'same_code')
    if factors['same_headers'] is not None and these_headers != factors['same_headers']:
        return ('http headers', params, 'same_headers')
    if mem.var['disable_redirects']:
        if factors['same_redirect'] is not None and urlparse(response.headers.get('Location', '')).path != factors['same_redirect']:
            return ('redirection', params, 'same_redirect')
    elif factors['same_redirect'] is not None and 'Location' in response.headers:
        if urlparse(response.headers.get('Location', '')).path != factors['same_redirect']:
            return ('redirection', params, 'same_redirect')
    if factors['same_body'] is not None and response.text != factors['same_body']:
        return ('body length', params, 'same_body')
    if factors['lines_num'] is not None and response.text.count('\n') != factors['lines_num']:
        return ('number of lines', params, 'lines_num')
    if factors['same_plaintext'] is not None and remove_tags(response.text) != factors['same_plaintext']:
        return ('text length', params, 'same_plaintext')
    if factors['lines_diff'] is not None:
        for line in factors['lines_diff']:
            if line not in response.text:
                return ('lines', params, 'lines_diff')
    if factors['param_missing'] is not None:
        for param in params.keys():
            if len(param) < 5:
                continue
            if param not in factors['param_missing'] and re.search(r'[\'"\s]%s[\'"\s]' % re.escape(param), response.text):
                return ('param name reflection', params, 'param_missing')
    if factors['value_missing'] is not None:
        for value in params.values():
            if type(value) != str or len(value) != 6:
                continue
            if value in response.text and re.search(r'[\'"\s]%s[\'"\s]' % re.escape(value), response.text):
                return ('param value reflection', params, 'value_missing')
    return ('', [], '')



import requests

from urllib.parse import urlparse


def commoncrawl(host, page=0):
	these_params = set()
	response = requests.get('http://index.commoncrawl.org/CC-MAIN-2024-42-index?url=*.%s&fl=url&page=%s&limit=10000' % (host, page), verify=False).text
	if response.startswith('<!DOCTYPE html>'):
		return ([], False, 'commoncrawl')
	urls = response.split('\n')
	for url in urls:
		for param in urlparse(url).query.split('&'):
			these_params.add(param.split('=')[0])
	return (these_params, True, 'commoncrawl')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
from setuptools import setup, find_packages
from os import path
this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    desc = f.read()

setup(
    name='arjun',
    version=__import__('arjun').__version__,
    description='HTTP parameter discovery suite',
    long_description=desc,
    long_description_content_type='text/markdown',
    author='Somdev Sangwan',
    author_email='s0md3v@gmail.com',
    license='GNU General Public License v3 (GPLv3)',
    url='https://github.com/s0md3v/Arjun',
    download_url='https://github.com/s0md3v/Arjun/archive/v%s.zip' % __import__('arjun').__version__,
    zip_safe=False,
    packages=find_packages(),
    package_data={'arjun': ['db/*']},
    install_requires=[
        'requests',
        'dicttoxml',
        'ratelimit'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Operating System :: OS Independent',
        'Topic :: Security',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3.4',
    ],
    entry_points={
        'console_scripts': [
            'arjun = arjun.__main__:main'
        ]
    },
    keywords=['arjun', 'bug bounty', 'http', 'pentesting', 'security'],
)



import os
import tempfile


def prompt(default=None):
    """
    lets user paste input by opening a temp file in a text editor
    returns str (content of tmp file)
    """
    editor = 'nano'
    with tempfile.NamedTemporaryFile(mode='r+') as tmpfile:
        if default:
            tmpfile.write(default)
            tmpfile.flush()

        child_pid = os.fork()
        is_child = child_pid == 0

        if is_child:
            os.execvp(editor, [editor, tmpfile.name])
        else:
            os.waitpid(child_pid, 0)
            tmpfile.seek(0)
            return tmpfile.read().strip()



import re

burp_regex = re.compile(r'''(?m)^    <url><!\[CDATA\[(.+?)\]\]></url>
    <host ip="[^"]*">[^<]+</host>
    <port>[^<]*</port>
    <protocol>[^<]*</protocol>
    <method><!\[CDATA\[(.+?)\]\]></method>
    <path>.*</path>
    <extension>(.*)</extension>
    <request base64="(?:false|true)"><!\[CDATA\[([\s\S]+?)]]></request>
    <status>([^<]*)</status>
    <responselength>([^<]*)</responselength>
    <mimetype>([^<]*)</mimetype>''')


def reader(path, mode='string'):
    """
    reads a file
    returns a string/array containing the content of the file
    """
    with open(path, 'r', encoding='utf-8') as file:
        if mode == 'lines':
            return list(filter(None, [line.rstrip('\n') for line in file]))
        else:
            return ''.join([line for line in file])


def parse_request(string):
    """
    parses http request
    returns dict
    """
    result = {}
    match = re.search(r'(?:([a-zA-Z0-9]+) ([^ ]+) [^ ]+\n)?([\s\S]+\n)\n?([\s\S]+)?', string)
    result['method'] = match.group(1)
    result['path'] = match.group(2)
    result['headers'] = parse_headers(match.group(3))
    result['url'] = 'http://' + result['headers']['Host'] + result['path']
    result['data'] = match.group(4)
    return result


def parse_headers(string):
    """
    parses headers
    return dict
    """
    result = {}
    for line in string.split('\n'):
        if len(line) > 1:
            splitted = line.split(':')
            result[splitted[0]] = ':'.join(splitted[1:]).strip()
    return result


def burp_import(path):
    """
    imports targets from burp suite
    returns list (of request objects)
    """
    requests = []
    content = reader(path)
    matches = re.finditer(burp_regex, content)
    for match in matches:
        request = parse_request(match.group(4))
        headers = request['headers']
        if match.group(7) in ('HTML', 'JSON'):
            requests.append({
                'url': match.group(1),
                'method': match.group(2),
                'extension': match.group(3),
                'headers': headers,
                'include': request['data'],
                'code': match.group(5),
                'length': match.group(6),
                'mime': match.group(7)
            })
    return requests


def urls_import(path, method, headers, include):
    """
    imports urls from a newline delimited text file
    returns list (of request objects)
    """
    requests = []
    urls = reader(path, mode='lines')
    for url in urls:
        requests.append({
            'url': url,
            'method': method,
            'headers': headers,
            'data': include
        })
    return requests


def request_import(path):
    """
    imports request from a raw request file
    returns list
    """
    result = []
    result.append(parse_request(reader(path)))
    return result


def importer(path, method, headers, include):
    """
    main importer function that calls other import functions
    """
    with open(path, 'r', encoding='utf-8') as file:
        for line in file:
            if line.startswith('<?xml'):
                return burp_import(path)
            elif line.startswith(('http://', 'https://')):
                return urls_import(path, method, headers, include)
            elif line.startswith(('GET', 'POST')):
                return request_import(path)
            return []



import json
import requests

import arjun.core.config as mem
from arjun.core.utils import populate

from arjun.core.utils import create_query_string


def json_export(result):
    """
    exports result to a file in JSON format
    """
    with open(mem.var['json_file'], 'w+', encoding='utf8') as json_output:
        json.dump(result, json_output, sort_keys=True, indent=4)


def burp_export(result):
    """
    exports results to Burp Suite by sending request to Burp proxy
    """
    proxy = ('' if ':' in mem.var['burp_proxy'] else '127.0.0.1:') + mem.var['burp_proxy']
    proxies = {
        'http': 'http://' + proxy,
        'https': 'https://' + proxy
    }
    for url, data in result.items():
        if data['method'] == 'GET':
            requests.get(url, params=populate(data['params']), headers=data['headers'], proxies=proxies, verify=False)
        elif data['method'] == 'POST':
            requests.post(url, data=populate(data['params']), headers=data['headers'], proxies=proxies, verify=False)
        elif data['method'] == 'JSON':
            requests.post(url, json=populate(data['params']), headers=data['headers'], proxies=proxies, verify=False)


def text_export(result):
    """
    exports results to a text file, one url per line
    """
    with open(mem.var['text_file'], 'a+', encoding='utf8') as text_file:
        for url, data in result.items():
            clean_url = url.lstrip('/')
            if data['method'] == 'JSON':
                text_file.write(clean_url + '\t' + json.dumps(populate(data['params'])) + '\n')
            else:
                query_string = create_query_string(data['params'])
                if '?' in clean_url:
                    query_string = query_string.replace('?', '&', 1)
                if data['method'] == 'GET':
                    text_file.write(clean_url + query_string + '\n')
                elif data['method'] == 'POST':
                    text_file.write(clean_url + '\t' + query_string + '\n')


def exporter(result):
    """
    main exporter function that calls other export functions
    """
    if mem.var['json_file']:
        json_export(result)
    if mem.var['text_file']:
        text_export(result)
    if mem.var['burp_proxy']:
        burp_export(result)



import re
import sys
import json
import random
import requests

import concurrent.futures
from dicttoxml import dicttoxml
from urllib.parse import urlparse

from arjun.core.prompt import prompt
from arjun.core.importer import importer

from arjun.plugins.otx import otx
from arjun.plugins.wayback import wayback
from arjun.plugins.commoncrawl import commoncrawl

import arjun.core.config as mem
from arjun.core.colors import info


def extract_headers(headers):
    """
    parses headers provided through command line
    returns dict
    """
    headers = headers.replace('\\n', '\n')
    return parse_headers(headers)


def confirm(array_of_dicts, usable):
    """
    extracts the value from single valued dict from an array of dicts
    returns an array of dicts
    """
    param_groups = []
    for dic in array_of_dicts:
        if len(dic) == 1:
            usable.append(dic)
        else:
            param_groups.append(dic)
    return param_groups


def slicer(dic, n=2):
    """
    divides dict into n parts
    returns array containing n dicts
    """
    listed = list(dic.items())
    k, m = divmod(len(dic), n)
    return [dict(listed[i * k + min(i, m):(i + 1) * k + min(i + 1, m)]) for i in range(n)]


def populate(array):
    """
    converts a list of parameters into parameter and value pair
    returns dict
    """
    return {name: '1' * (6 - len(str(i))) + str(i) for i, name in enumerate(array)}


def stable_request(url, headers):
    """
    guarantees crash-proof HTTP(S) requests
    returns None in case of failure, returns a "response" object otherwise
    """
    parsed = urlparse(url)
    redirects_allowed = False if mem.var['disable_redirects'] else True
    scheme, host, path = parsed.scheme, parsed.netloc, parsed.path
    schemes = (['https', 'http'] if scheme == 'https' else ['http', 'https'])
    for scheme in schemes:
        try:
            response = requests.get(
                scheme + '://' + host + path,
                headers=headers,
                verify=False,
                timeout=10,
                allow_redirects=redirects_allowed)
            content = response.headers.get('Content-Type', '')
            if not ('text' in content or 'html' in content or 'json' in content or 'xml' in content):
                print('%s URL doesn\'t seem to be a webpage. Skipping.' % info)
                return None
            return response.url
        except Exception as e:
            if 'ConnectionError' not in str(e):
                continue
        return None


def remove_tags(html):
    """
    removes all the html from a webpage source
    """
    return re.sub(r'(?s)<.*?>', '', html)


def diff_map(body_1, body_2):
    """
    creates a list of lines that are common between two multi-line strings
    returns list
    """
    sig = []
    lines_1, lines_2 = body_1.split('\n'), body_2.split('\n')
    for line_1, line_2 in zip(lines_1, lines_2):
        if line_1 == line_2:
            sig.append(line_1)
    return sig


def random_str(n):
    """
    generates a random string of length n
    returns a string containing only digits
    """
    return ''.join(str(random.choice(range(10))) for i in range(n))


def get_params(include):
    """
    loads parameters from JSON/query string
    returns parameter dict
    """
    params = {}
    if include:
        if include.startswith('{'):
            try:
                params = json.loads(str(include).replace('\'', '"'))
                if type(params) != dict:
                    return {}
                return params
            except json.decoder.JSONDecodeError:
                return {}
        else:
            cleaned = include.split('?')[-1]
            parts = cleaned.split('&')
            for part in parts:
                each = part.split('=')
                try:
                    params[each[0]] = each[1]
                except IndexError:
                    params = {}
    return params


def create_query_string(params):
    """
    creates a query string from a list of parameters
    returns str
    """
    query_string = ''
    for param in params:
        pair = param + '=' + random_str(4) + '&'
        query_string += pair
    if query_string.endswith('&'):
        query_string = query_string[:-1]
    return '?' + query_string


def reader(path, mode='string'):
    """
    reads a file
    returns a string/array containing the content of the file
    """
    with open(path, 'r', encoding='utf-8') as file:
        if mode == 'lines':
            return list(filter(None, [line.rstrip('\n') for line in file]))
        else:
            return ''.join([line for line in file])


def extract_js(response):
    """
    extracts javascript from a given string
    """
    scripts = []
    for part in re.split('(?i)<script[> ]', response):
        actual_parts = re.split('(?i)</script>', part, maxsplit=2)
        if len(actual_parts) > 1:
            scripts.append(actual_parts[0])
    return scripts


def parse_headers(string):
    """
    parses headers
    returns dict
    """
    result = {}
    for line in string.split('\n'):
        if len(line) > 1:
            splitted = line.split(':')
            result[splitted[0]] = ':'.join(splitted[1:]).strip()
    return result


def parse_request(string):
    """
    parses http request
    returns dict
    """
    result = {}
    match = re.search(r'(?:([a-zA-Z0-9]+) ([^ ]+) [^ ]+\n)?([\s\S]+\n)\n?([\s\S]+)?', string)
    result['method'] = match.group(1)
    result['path'] = match.group(2)
    result['headers'] = parse_headers(match.group(3))
    result['data'] = match.group(4)
    return result


def http_import(path):
    """
    parses http request from a file
    returns dict
    """
    return parse_request(reader(path))


def fetch_params(host):
    """
    fetch parameters from passive sources
    returns list
    """
    available_plugins = {'commoncrawl': commoncrawl, 'otx': otx, 'wayback': wayback}
    page = 0
    progress = 0
    params = {}
    while len(available_plugins) > 0 and page <= 10:
        threadpool = concurrent.futures.ThreadPoolExecutor(max_workers=len(available_plugins))
        futures = (threadpool.submit(func, host, page) for func in available_plugins.values())
        for each in concurrent.futures.as_completed(futures):
            if progress < 98:
                progress += 3
            this_result = each.result()
            if not this_result[1]:
                progress += ((10 - page) * 10 / 3)
                del available_plugins[this_result[2]]
            if len(this_result[0]) > 1:
                if not params:
                    params = this_result[0]
                else:
                    params.update(this_result[0])
            print('%s Progress: %i%%' % (info, progress), end='\r')
        page += 1
    print('%s Progress: %i%%' % (info, 100), end='\r')
    return params


def prepare_requests(args):
    """
    creates a list of request objects used by Arjun from targets given by user
    returns list (of targets)
    """
    headers = {
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:83.0) Gecko/20100101 Firefox/83.0',
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'Accept-Language': 'en-US,en;q=0.5',
        'Accept-Encoding': 'gzip, deflate',
        'Connection': 'close',
        'Upgrade-Insecure-Requests': '1'
    }
    result = []
    if type(args.headers) == str:
        headers = extract_headers(args.headers)
    elif args.headers:
        headers = extract_headers(prompt())
    if mem.var['method'] == 'JSON':
        headers['Content-type'] = 'application/json'
    if args.url:
        params = get_params(args.include)
        result.append(
            {
                'url': args.url,
                'method': mem.var['method'],
                'headers': headers,
                'include': params
            }
        )
    elif args.import_file:
        result = importer(args.import_file, mem.var['method'], headers, args.include)
    return result


def nullify(*args, **kwargs):
    """
    a function that does nothing
    """
    pass


def dict_to_xml(dict_obj):
    """
    converts dict to xml string
    returns str
    """
    return dicttoxml(dict_obj, root=False, attr_type=False).decode('utf-8')


def compatible_path(path):
    """
    converts filepaths to be compatible with the host OS
    returns str
    """
    if sys.platform.lower().startswith('win'):
        return path.replace('/', '\\')
    return path
